@@ -10,9 +10,10 @@ Date: April 5th, 2021.
 import json
 import logging
 import os
+from concurrent.futures import ThreadPoolExecutor
 from functools import wraps
 
-from oauthlib.oauth2 import BackendApplicationClient
+from oauthlib.oauth2 import BackendApplicationClient, OAuth2Token
 from requests_oauthlib import OAuth2Session
 
 from . import status
@@ -50,7 +51,9 @@ from .utils import sentinel
 try:
     from django.core.cache import cache
 except ImportError:
-    from .utils import cache
+    from .utils import SQLiteCache
+
+    cache = SQLiteCache()
 
 
 __all__ = ["DynamicsClient"]
@@ -59,6 +62,30 @@ __all__ = ["DynamicsClient"]
 logger = logging.getLogger(__name__)
 
 
+def get_token() -> OAuth2Token:
+    """Get dynamics client token in a thread, so it can be done in an async context."""
+
+    def task() -> OAuth2Token:
+        return cache.get("dynamics-client-token", None)
+
+    with ThreadPoolExecutor() as executor:
+        future = executor.submit(task)
+        return future.result()
+
+
+def set_token(token: OAuth2Token):
+    """Set dynamics client token in a thread, so it can be done in an async context."""
+
+    def task():
+        name = "dynamics-client-token"
+        expires = int(token["expires_in"]) - 60
+        cache.set(name, token, expires)
+
+    with ThreadPoolExecutor() as executor:
+        future = executor.submit(task)
+        return future.result()
+
+
 def error_simplification_available(func):
     """Errors in the function decorated with this decorator can be simplified to just a
     DynamicsException with default error message using the keyword: 'simplify_errors'.
@@ -81,7 +108,8 @@ def error_simplification_available(func):
             logger.warning(error)
             if not simplify_errors or any(isinstance(error, exception) for exception in raise_separately):
                 raise error
-            raise DynamicsException("There was a problem communicating with the server.") from error
+            self: "DynamicsClient" = args[0]
+            raise DynamicsException(self.simplified_error_message) from error
 
     return inner
 
@@ -90,6 +118,10 @@ class DynamicsClient:  # pylint: disable=R0904,R0902
     """Dynamics client for making queries from a Microsoft Dynamics 365 database."""
 
     request_counter: int = 0
+    actions_class: Type[Actions] = Actions
+    functions_class: Type[Functions] = Functions
+    simplified_error_message: str = "There was a problem communicating with the server."
+
     error_dict = {
         status.HTTP_400_BAD_REQUEST: ParseError,
         status.HTTP_401_UNAUTHORIZED: AuthenticationFailed,
@@ -99,6 +131,7 @@ class DynamicsClient:  # pylint: disable=R0904,R0902
         status.HTTP_412_PRECONDITION_FAILED: DuplicateRecordError,
         status.HTTP_413_REQUEST_ENTITY_TOO_LARGE: PayloadTooLarge,
         status.HTTP_429_TOO_MANY_REQUESTS: APILimitsExceeded,
+        status.HTTP_500_INTERNAL_SERVER_ERROR: DynamicsException,
         status.HTTP_501_NOT_IMPLEMENTED: OperationNotImplemented,
         status.HTTP_503_SERVICE_UNAVAILABLE: WebAPIUnavailable,
     }
@@ -119,16 +152,16 @@ class DynamicsClient:  # pylint: disable=R0904,R0902
 
         self._api_url = api_url.rstrip("/") + "/"
         self._session = OAuth2Session(client=BackendApplicationClient(client_id=client_id))
-        token = cache.get("dynamics-client-token", None)
+        token = get_token()
         if token is None:
             token = self._session.fetch_token(token_url=token_url, client_secret=client_secret, scope=scope)
-            cache.set("dynamics-client-token", token, int(token.get("expires_in", 3599)) - 60)
+            set_token(token)
         else:
             self._session.token = token
 
-        self.actions = Actions(client=self)
+        self.actions = self.actions_class(client=self)
         """Predefined dynamics API actions. Used to make certain changes to data."""
-        self.functions = Functions(client=self)
+        self.functions = self.functions_class(client=self)
         """Predefined dynamics API functions. Used to fetch certain types of data."""
 
         self._select: List[str] = []
@@ -167,13 +200,13 @@ class DynamicsClient:  # pylint: disable=R0904,R0902
         :raises KeyError: An environment variable was not configured properly
         """
 
-        base_url = os.environ["DYNAMICS_API_URL"]
+        api_url = os.environ["DYNAMICS_API_URL"]
         token_url = os.environ["DYNAMICS_TOKEN_URL"]
         client_id = os.environ["DYNAMICS_CLIENT_ID"]
         client_secret = os.environ["DYNAMICS_CLIENT_SECRET"]
         scope = os.environ["DYNAMICS_SCOPE"].split(",")
 
-        return cls(base_url, token_url, client_id, client_secret, scope)
+        return cls(api_url, token_url, client_id, client_secret, scope)
 
     @property
     def current_query(self) -> str:
@@ -263,21 +296,23 @@ class DynamicsClient:  # pylint: disable=R0904,R0902
 
         self.headers.setdefault("Prefer", f"odata.maxpagesize={self.pagesize}")
 
-    def _error_handling(self, status_code: int, error_message: str, method: MethodType):
+    def error_handling(self, status_code: int, error_message: str, error_code: str, method: MethodType):
         """Error handling based on these expected error statuses:
         https://docs.microsoft.com/en-us/powerapps/developer/data-platform/webapi/compose-http-requests-handle-errors#identify-status-codes
 
         :param status_code: Error status from dynamics
         :param error_message: Error message from dynamics.
+        :param error_code: Error code from dynamics.
         :param method: HTTP method in question.
         """
 
         logger.warning(
-            "Dynamics client <[%s] %s> failed with status %d: %s",
+            "Dynamics client <[%s] %s> failed with status %d: %s (%s)",
             method.upper(),
             self.current_query,
             status_code,
             error_message,
+            error_code,
         )
         error = self.error_dict.get(status_code, DynamicsException)
         raise error(detail=error_message)
@@ -308,12 +343,22 @@ class DynamicsClient:  # pylint: disable=R0904,R0902
         count = data.get("@odata.count", "")
 
         if errors:
-            self._error_handling(status_code=response.status_code, error_message=errors.get("message"), method="get")
+            self.error_handling(
+                status_code=response.status_code,
+                error_message=errors.get("message"),
+                error_code=errors.get("code"),
+                method="get",
+            )
         elif not entities:
             if not_found_ok:
                 return []
             message = "No records matching the given criteria."
-            self._error_handling(status_code=status.HTTP_404_NOT_FOUND, error_message=message, method="get")
+            self.error_handling(
+                status_code=status.HTTP_404_NOT_FOUND,
+                error_message=message,
+                error_code="not_found",
+                method="get",
+            )
 
         # Fetch more data if needed
         for i, row in enumerate(entities):
@@ -369,7 +414,12 @@ class DynamicsClient:  # pylint: disable=R0904,R0902
 
         errors = data.get("error")
         if errors:
-            self._error_handling(status_code=response.status_code, error_message=errors.get("message"), method="post")
+            self.error_handling(
+                status_code=response.status_code,
+                error_message=errors.get("message"),
+                error_code=errors.get("code"),
+                method="post",
+            )
 
         return data
 
@@ -396,7 +446,12 @@ class DynamicsClient:  # pylint: disable=R0904,R0902
 
         errors = data.get("error")
         if errors:
-            self._error_handling(status_code=response.status_code, error_message=errors.get("message"), method="patch")
+            self.error_handling(
+                status_code=response.status_code,
+                error_message=errors.get("message"),
+                error_code=errors.get("code"),
+                method="patch",
+            )
 
         return data
 
@@ -419,7 +474,12 @@ class DynamicsClient:  # pylint: disable=R0904,R0902
 
         errors = data.get("error")
         if errors:
-            self._error_handling(status_code=response.status_code, error_message=errors.get("message"), method="delete")
+            self.error_handling(
+                status_code=response.status_code,
+                error_message=errors.get("message"),
+                error_code=errors.get("code"),
+                method="delete",
+            )
 
     @property
     def table(self) -> str:
@@ -751,3 +811,9 @@ class DynamicsClient:  # pylint: disable=R0904,R0902
             raise ValueError(f"Max pagesize is 5000. Got {value}.")
 
         self._pagesize = value
+
+    # TODO: fetchXml:
+    #  https://docs.microsoft.com/en-us/powerapps/developer/data-platform/
+    #   use-fetchxml-construct-query
+    #  https://docs.microsoft.com/en-us/powerapps/developer/data-platform/
+    #   webapi/retrieve-and-execute-predefined-queries#use-custom-fetchxml
