@@ -51,6 +51,12 @@ class DynamicsException(APIException):
     default_detail = "Dynamics Web API call failed."
     default_code = "dynamics_link_failed"
 
+    def __init__(self, detail=None, code=None, **kwargs):  # pylint: disable=W0613
+        detail = self.default_detail if detail is None else detail
+        code = self.default_code if code is None else code
+        self.args = (str(detail),)  # since APIException doesn't call super()
+        super().__init__(detail, code)
+
 
 class ParseError(DynamicsException):
     status_code = status.HTTP_400_BAD_REQUEST
